import numpy as np
import pandas as pd
from Bio import PDB, pairwise2
from Bio.PDB.DSSP import DSSP

from copy import copy
import subprocess
import json

def readPDB(filename, name=None, chains='A', model=0, remove_unknown=True, fmt="detect"):

    '''
    Given a PDB code and a directory of where to find the structure, returns
    the distance matrix and residue ids.

    Inputs:
        filename: the filepath to your .pdb or .cif file
        name: an optional name to assign the structure, can be the PDB code
            (generated automatically from the filename if you don't pass one)
        chains: which chains of the structure to load. Defaults to "A."
            Can also pass 'all', which will load all chains.
        model: which model to use from multi-model PDBs. Defaults to 0."

        remove_unknown: filter out residues with identifier "UNK" (unknown), present
            in some pdbs, which can lead to errors downstream
            (defaults to TRUE because I assume such "residues" are rarely useful, but
             you can turn it off if you want to analyze the unknown residues)

    Outputs:
        structure: a ProteinStructure object
    '''

    # If you don't pass a name, automatically fill in a name from the pdb file
    #  (for some reason Biopython's PDBParser requires a "PDB code" which can really
    #    just be any name, doesn't have to be the real PDB code)
    if type(name) == type(None):
        name = filename.split(".")[0]
        
    if fmt=="detect":
        fmt = filename.split('.')[-1]

    if fmt.lower()=='pdb':
        # Load the structure using Bio.PDB
        structure = PDB.PDBParser(QUIET=True).get_structure(name, filename)
        
    elif fmt.lower()=="cif" or fmt.lower()=="mmcif":
        structure = PDB.MMCIFParser(QUIET=True).get_structure(name, filename)

    # The number model to use in the structure; usually there is only one, and this is 0 (default)
    m = structure[model]

    # Which chains to extract - either all chains, or a single chain
    if isinstance(chains, str):
        if chains.lower()=='all':    # Just extract all chains
            structure_list = []
            for ch in m:
                structure_list.append(ProteinStructure(ch, name=name+'-'+ch.id, remove_unknown=remove_unknown))
            return structure_list

        else:    # If it's a string and not 'All', we assume it's a single chain index
            try:    # A try... except here so that it can tell you if you put in an incorrect chain ID
                ch = m[chains]
            except:
                print(chains, 'is not a valid chain ID in', filename)
                return None

            return ProteinStructure(ch, name=name, remove_unknown=remove_unknown)

    else:
        # If you pass it a list, it will extract those chains
        structure_list = []
        for ch in m:
            structure_list.append(ProteinStructure(ch, name=name+'-'+ch.id, remove_unknown=remove_unknown))
        return structure_list


def readAlphaFold(filename, name, pae_file=None):

    structure = Bio.PDB.PDBParser(QUIET=True).get_structure(name, filename)
    prot = ProteinStructure(structure[0]["A"], name=name)

    if type(pae_file)!=type(None):

        pae_matrix = readPAE(pae_file)



        prot.pae = pae_matrix

    return prot


letters = {'ALA':'A','ARG':'R','ASN':'N','ASP':'D','CYS':'C','GLU':'E','GLN':'Q','GLY':'G','HIS':'H',
           'ILE':'I','LEU':'L','LYS':'K','MET':'M','PHE':'F','PRO':'P','SER':'S','THR':'T','TRP':'W',
           'TYR':'Y','VAL':'V', 'MSE':'M', 'UNK':'X'}

class ProteinStructure:

    '''
    A class for protein structures, which uses Biopython residue and atom objects but is intended to
    be much easier to use that a Biopython <structure> object, which are just a complete mess.
    '''

    def __init__(self, chain, annotation_type='bfactor', annotate_by='residue', resolution=None, name=None,
                remove_unknown=True):

        self.chain = chain

        if remove_unknown:
            self.residues = [residue for residue in chain if residue.resname != "UNK"]
        else:
            self.residues = [residue for residue in chain]
        self.residue_ids = np.array([residue.id[1] for residue in self.residues])
        self.sequence = ''.join([letters[residue.resname] for residue in self.residues if residue.resname in letters])
        self.ordered_sequence = ''.join([letters[residue.resname] for residue in self.residues if residue.resname in letters and not residue.is_disordered()])

        self.n_residues = len(self.residues)
        self.atoms = np.concatenate([[atom for atom in residue] for residue in self.residues if not residue.is_disordered()])
        self.n_atoms = len(self.atoms)
        self.xyz = np.array([atom.coord for atom in self.atoms])
        self.dmatrix = None; self.distance_method = None
        self._annotation_type = annotation_type
        self._annotate_by = annotate_by
        self.resolution = resolution
        self.name = name
        if len(self.name) == 4:
            self.pdb = self.name.upper()

        if annotation_type.lower()=='bfactor':

            if 'res' in annotate_by.lower():
                bfactors = []
                for residue in self.residues:
                    bfactors.append(np.mean([atom.bfactor for atom in residue]))
                self.bfactors = np.array(bfactors)
            else:
                self.bfactors = np.array([atom.bfactor for atom in self.atoms])

        elif annotation_type.lower()=='plddm':

            plddms = []
            for residue in self.residues:
                plddms.append(np.max([atom.bfactor for atom in residue]))

            self.plddm = np.array(plddms)

    def __len__(self):
        return self.n_residues

    def __str__(self):
        return "Protein structure "+self.name+" with "+str(self.n_residues)+" residues and "+str(self.n_atoms)+" atoms"

    def __getitem__(self, index):

        if isinstance(index, int):
            try:
                ind = int(np.where(self.residue_ids==index)[0][0])
                return self.residues[ind]
            except:
                return ValueError("No such residue in protein!")

        elif isinstance(index, slice):
            return ProteinStructure(self.residues[index], name=self.name,
                    annotation_type=self._annotation_type, annotate_by=self._annotate_by,
                    resolution=self.resolution)

        elif isinstance(index, (list, np.ndarray)):
            return ProteinStructure([self.residues[k] for k in index], name=self.name,
                    annotation_type=self._annotation_type, annotate_by=self._annotate_by,
                    resolution=self.resolution)

    def __iter__(self):
        return iter(self.residues)

    def select_atoms(self, key):

        if isinstance(key, str):

            # if ' or ' in str:
            #     as = []
            #     for x in key.split(' or '):
            #         as.append([n for n,atom in enumerate(self.atoms) if atom.name==x])
            #     return self.select_atoms(np.concatenate(as))

            atom_sel = [n for n,atom in enumerate(self.atoms) if atom.name==key]
            return self.select_atoms(atom_sel)

        elif isinstance(key, int):
            return self.atoms[key]

        else:
            residues_in_selection = [self.atoms[k].get_parent().id[1] for k in key]
            unique_residues = np.unique(residues_in_selection)
            where_res = [np.where(np.array(residues_in_selection)==res)[0] for res in unique_residues]

            final_residue_list = []

            for n,resn in enumerate(unique_residues):
                res = self.residues[list(self.residue_ids).index(resn)]
                newres = PDB.Residue.Residue(res.id, res.resname, res.segid)

                for atom_index in where_res[n]:
                    newres.add(copy(self.atoms[key[atom_index]]))

                final_residue_list.append(newres)

            return ProteinStructure(final_residue_list, annotation_type=self._annotation_type,
                                    annotate_by=self._annotate_by, resolution=self.resolution, name=self.name)


    def distance_matrix(self, method="CA", recalculate=False):

        if type(self.dmatrix)==type(None) or recalculate==True:

            if method=="CA":
                self.dmatrix= CalcDistanceMatrix(self.select_atoms("CA"))
                self.distance_method = "CA"
            elif method.lower() in ('full', 'all'):
                self.dmatrix = CalcDistanceMatrix(self)
                self.distance_method = "full"

        elif self.distance_method != method:
            if method=="CA":
                self.dmatrix= CalcDistanceMatrix(self.select_atoms("CA"))
                self.distance_method = "CA"
            elif method.lower() in ('full', 'all'):
                self.dmatrix = CalcDistanceMatrix(self)
                self.distance_method = "full"

        return self.dmatrix

    def distance(self, i, j, method="CA", atoms=None):

        '''
        Get the distance between two residues in a structure by passing the numbers of two residues.

        Has three available 'modes' of calculating the distance:

            "C-alphas" : gives the distance between the alpha carbons of each residue (DEFAULT)
                           set method="CA", atoms=None (Default)
            "min" :      minimum distance between any atoms in the residues
                           set method="min", atoms=None (Default)
            "atoms" :      pick user-defined atoms from each residue to get a very specific distance
                            atoms = (ID_1, ID_2), e.g. atoms=("O1", "O2") gives distance between O1 in the
                            first residue and O2 in the second residue.
                            the method keyword can be set to anything, it is ignored

        '''

        if isinstance(atoms, (list, tuple)):

            if len(atoms)==2:
                a1 = ProteinStructure([structure_1.residues[i]], name='p1').select(atoms(atoms[0])).xyz[0]
                a2 = ProteinStructure([structure_1.residues[j]], name='p2').select(atoms(atoms[1])).xyz[0]

                return np.sqrt(np.sum((a1 - a2)**2))

            else:
                return IndexError(
                    "To get distances between user-selected atoms, you must pass a list of 2 valid PDB atom indices"
                )

        elif method.upper() == 'CA':
            calphas = self.select_atoms("CA")
            return np.sqrt(np.sum((calphas.xyz[i]-calphas.xyz[j])**2))

        elif method.upper() == 'MIN':
            a1 = ProteinStructure([structure_1.residues[i]], name='p1').xyz
            a2 = ProteinStructure([structure_1.residues[j]], name='p2').xyz
            return np.min(np.sqrt(np.sum((a1[:,None,:] - a2[None,:,:])**2, axis=2)))

    def contacts(self, threshold, dist_threshold=None):

        return get_contacts(self, threshold, dist_threshold)

    def annotate_secondary_structure(dssp_path='mkdssp', filename='temp.pdb'):

        self.save(self, filename)
        self.ss = PDB.DSSP.DSSP(chain, filename, dssp_path)

    def ordered_residues(self):
        return [res for res in self.residues if not res.is_disordered()]

    def save(self, pdb_file):
        io = PDB.PDBIO()
        io.set_structure(self.chain)
        io.save(pdb_file)

#class Complex(Structure):

def get_contacts(structure, thresh, dist_threshold=None):

    '''
    Calculate contacts for which the C-alphas are within some threshold for a ProteinStructure object.
    '''

    if type(structure.dmatrix)==type(None):
        structure.dmatrix = CalcDistanceMatrix(structure)

    if type(dist_threshold)==type(None):

        selector = np.where(structure.distance_matrix() < thresh)

    else:
        s = structure.dmatrix.shape[0]
        off_diag = np.abs(np.arange(s)[None,:] - np.arange(s)[:,None]) > dist_threshold
        selector = np.where((structure.distance_matrix() < thresh)&(off_diag))

    return np.array([structure.residue_ids[selector[0]], structure.residue_ids[selector[1]]]).T


def CalcDistanceMatrix(Structure):

    D = []

    for i in range(len(Structure.xyz)):
        D.append(np.sqrt(np.sum((Structure.xyz[i]-Structure.xyz)**2,axis=1)))

    return np.array(D)

def alphafold_from_uniprot(uniprot_id, version=4, pae=True, download_dir="./"):
    
    '''Use wget to download a particular AlphaFold structure directly from the database based on its UniProt ID.
    
    Input:
        uniprot_id: the UniProt ID of your sequence of interest, e.g. A0A812H9W4
        version: the version of AlphaFold that you want - as of writing, the newest is 4
        pae: whether to also download the PAE matrix file and attach is a .pae attribute'''
    
    alphafold_pdb = "AF-"+uniprot_id+"-F1-model_v"+str(version)+".pdb"
    subprocess.run("wget "+"https://alphafold.ebi.ac.uk/files/"+alphafold_pdb, shell=True)
    
    structure = readPDB(alphafold_pdb)
    
    if pae:
        pae_filename = "AF-"+uniprot_id+"-F1-predicted_aligned_error_v4.json"
        subprocess.run("wget "+"https://alphafold.ebi.ac.uk/files/"+pae_filename, shell=True)
        structure.pae = readPAE(pae_filename)
    
    return structure

def readPAE(filename):
    
    """Read a PAE json file from the AlphaFold database"""
    
    try:
        # For PAE files as a text file that can be loaded in directly
        pae_matrix = np.genfromtxt(pae_file)
    except:
        # Otherwise try the JSON method
        pae_matrix = np.array(json.load(open(filename ,"r"))[0]["predicted_aligned_error"])

    return pae_matrix



import numpy as np
import pandas as pd
from bioviper import msa
from Bio import Phylo
from io import StringIO
import ete3
import os
from copy import deepcopy
from matplotlib.colors import to_rgba, to_rgb, ListedColormap
from matplotlib.cm import get_cmap

from tqdm import tqdm

class Tree:

    def __init__(self, biopython_tree, tempfile=".temp.nwk", rooted=False):

        if isinstance(biopython_tree, str):
            biopython_tree = Phylo.read(StringIO(biopython_tree), format='newick')

        elif isinstance(biopython_tree, Phylo.Newick.Clade):
            self.branch_length = biopython_tree.branch_length
            biopython_tree = Phylo.Newick.Tree(biopython_tree)

        self._biopython = biopython_tree

        # Biopython's tree object is missing some easy way of just storing all of the terminals
        #  the iterator is honestly so annoying
        self.leaves = [term for term in self._biopython.get_terminals()]
        self.branches = [nonterm for nonterm in self._biopython.get_nonterminals()]

        # But OK you can also call them terminals and nonterminals if you want, I won't stop you
        self.nonterminals = self.branches
        self.terminals = self.leaves

        self.leaf_names = np.array([term.name for term in self._biopython.get_terminals()])
        self.N = len(self.leaf_names)
        self.ids = self.leaf_names
        self._tempfile = tempfile
        self.depths = self._biopython.depths

        # For compatibility with biopython workflows
        self.clade = self._biopython.clade

        if np.all([type(leaf.color)==type(None) for leaf in self.leaves]):
            self.is_colored = False
        else:
            self.is_colored = True
            self.colors = [leaf.color for leaf in self.leaves]

        #Phylo.write(self._biopython, self._tempfile, "newick")
        self.ete3 = ete3.Tree(self._biopython.__format__("newick"))

        self.rooted = rooted
        self.root = None

    def __getitem__(self, index):
        
        return Tree(self._biopython.clade[index])
    
    def __iter__(self):
        
        self.first_level_clades = [Tree(clade) for clade in self.clade]
        return iter(self.first_level_clades)
    
    def __repr__(self):
        
        return "bioviper.phylo.Tree with "+str(self.N)+" leaves"
    
    def __str__(self):

        return self._biopython.__format__("newick")

    def _update_ete3_tree(self):

        #Phylo.write(self._biopython, self._tempfile, "newick")
        #self.ete3 = ete3.Tree(self._tempfile)

        self.ete3 = ete3.Tree(self._biopython.__format__("newick"))

    def set_outgroup(self, outgroup, in_place=False, outgroup_is_ancestor=False):
    
        '''
        Root the tree with an outgroup. The outgroup may be either:
            (a) the name of a single node in the tree
            (b) a list of nodes, in which case we'll find their common ancestor and root with that node

            If doing (b), be careful that you understand what clae you're rooting with! Depending on the current
                structure of the tree, this common ancestor could be your current root, in which case the function
                will throw an error

        Optional arguments:
            in_place: wheter to modify the tree in place (TRUE) or return a new copy of the tree that is rooted (FALSE; default)
            outgroup_is_ancestor: whether to manually override and inform it that the outgroup given is already a common ancestor
                (used by the function to re-call itself recursively)
        '''

        if isinstance(root, (str, np.str_)) or outgroup_is_ancestor:

            if in_place:
                # this will change the original ete3 tree
                self.ete3.set_outgroup(root)
                self._biopython = Phylo.read(StringIO(self.ete3.write("newick")), "newick")

            else:
                # otherwise first copy the tree, modify the ete3 tree of that copy
                t_copy = deepcopy(self)
                t_copy.ete3.set_outgroup()

                T = phylo.Tree(t_copy.ete3.write("newick"))
                T.rooted = True
                T.root = T._biopython.root

                return T
            
        else:
            anc = t.ete3.get_common_ancestor(root)

            return self.get_outgroup(self, anc, in_place=in_place)

    def root_with_outgroup(self, outgroup, in_place=False):

        '''Simply calls set_outgroup(), for redundancy with biopython trees'''

        return self.set_outgroup(outgroup, in_place)


    def root_at_midpoint(self):

        self._biopython.root_at_midpoint()
        self.rooted = True
        self.root = self._biopython.root
        self._update_ete3_tree()

    def set_outgroup(self, outgroup):

        self._biopython.root_with_outgroup(outgroup)
        self.rooted = True
        self.root = self._biopython.root
        self._update_ete3_tree()
        self.ete3.set_outgroup(outgroup)

    def get_terminals(self):
        return self.leaves

    def get_leaf_index(self, name):

        x = np.where(self.leaf_names==name)[0]
        if len(x)==1:
            return x[0]
        elif len(x)==0:
            return []
        else:
            return x

    def get_leaf(self, name):

        return self.leaves[self.get_leaf_index(name)]

    def find_clades(self, target=None, terminal=None, order="preorder"):

        return self._biopython.find_clades(target=target, terminal=terminal,
            order=order)

    def search_leaf(self, name):

        if name in self.leaf_names:
            return self.get_leaf(name), self.get_leaf_index(name)

        else:

            leaves = []; leaf_indices = []

            for nl, leaf in enumerate(self.leaves):
                if name in leaf.name:
                    leaves.append(leaf)
                    leaf_indices.append(nl)

            if len(leaves)==1:
                return leaves[0], leaf_indices[0]

            else:
                return leaves, leaf_indices


    def prune(self, terms):

        pruned_ete3_tree = deepcopy(self.ete3)
        pruned_ete3_tree.prune([str(i) for i in terms])

        tr = Tree(Phylo.read(StringIO(pruned_ete3_tree.write("newick")), "newick"))

        if self.is_colored:

            tr.color([self.colors[self.get_leaf_index(i)] for i in tr.leaf_names])

        return tr

    def common_ancestor(self, leaves):

        try:
            return Tree(self._biopython.common_ancestor(leaves))

        except ValueError:
            return Tree(self._biopython.common_ancestor([self.search_leaf(leaf)[0].name for leaf in leaves]))

    def color(self, colors, cmap=None, color_branches=True, in_place=True, df_col="color",
                 branch_coloring_method="average", vmin=None, vmax=None, override_cmap=False):

        '''
        Color the tree based on a given input. There are a variety of things that you can pass for the "colors"
            argument:

            a list of array of colors as strings or
        '''

        self.is_colored = True

        # If you ask for it not in-place, first copy the tree, color that new one in place and return it
        if not in_place:
            NewTree = Tree(deepcopy(self._biopython))
            NewTree.color(colors)
            return NewTree

        else:

            # If you pass a colormap, interpret values as being among a color spectrum
            #   and run self.color_spectrum()
            #   override_cmap is used when color_spectrum() calls color again with the floats turned to RGBA
            if type(cmap) != type(None) and override_cmap==False:
                self.color_spectrum(colors, cmap=cmap, color_branches=color_branches, in_place=in_place, df_col=df_col,
                                   branch_coloring_method=branch_coloring_method, vmin=vmin, vmax=vmax)

            else:

                # If the colors argument is a list or array
                if isinstance(colors, (list, tuple, np.ndarray)):

                    # If it's the same length as the number of leaves in the tree, interpret
                    #  as corresponding to the leaves in the same order that they come in
                    if len(colors) == len(self.ids):

                        # Loop through and color each leaf
                        for leaf,color in zip(self.leaves, colors):
                            ColorLeaf(leaf, color)

                    # If colors is length 4, interpret as a constant RGBA value for the whole tree
                    elif len(colors)==4:
                        for leaf in self.leaves:
                            leaf.color = colors

                    # Otherwise it's uninterpretable, raise an error
                    else:
                        raise InputError("Cannot color tree, length of colors array must match number of terminals in tree")

                # If it's a dictionary, we can treat it similarly but we'll use the keys
                elif isinstance(colors, dict):
                    for key in dict:
                        ColorLeaf(get_leaf(key))

            # Pass a pandas dataframe? Not implemented yet
            #elif isinstance(colors, pd.DataFrame):

                #colors_filtered = colors.loc[self.leaf_names]

                #for key,color in colors_filtered[df_col]:
                    #ColorLeaf(self.get_leaf())

                elif isinstance(colors, (str, np.str_)):
                    for leaf in self.leaves:
                        leaf.color = to_branch_color(to_rgb(colors))

                if color_branches:
                    self.recolor_branches(method=branch_coloring_method)

        self.colors = [leaf.color for leaf in self.leaves]

    def color_spectrum(self, color_arr, cmap='viridis', vmin=None, vmax=None, **kwargs):


        '''Color the tree using numeric values using a colormap. First generates colors from the colormap
            and then runs self.color() with those colors.

            Inputs:
                color_arr: an array of the same length - and same order - as self.leaves
                    containining floats or integers
                    (can also be a dictionary where keys are leaves in the tree)

                cmap: a valid matplotlib colormap, as a string or the actual colormap object
                vmin: the minimum value to set to 0 on the colormap. defaults to the minimum of the data
                vmax: the maximum value to set to 1 on the colormap. defaults to the maximum of the data
                other arguments pass through to self.color(), see above for those'''

        if isinstance(color_arr, (list, tuple)):
            color_arr = np.array(color_arr)

        if isinstance(cmap, (str, np.str_)):
            cmap = get_cmap(cmap)

        if type(vmin)==type(None):
            vmin = np.min(color_arr)

        if type(vmax)==type(None):
            vmax = np.max(color_arr)

        if isinstance(color_arr, dict):

            colors = {}

            for key, val in color_arr.items():
                colors[key] = cmap((val - vmin)/(vmax - vmin))
        else:

            normed_color_arr = (color_arr - vmin)/(vmax-vmin)
            colors = cmap(normed_color_arr)

        return self.color(colors, override_cmap=True, **kwargs)


    def recolor_branches(self, method="average", default_color='black'):

        '''
        Recolor internal branches of the tree based on the colors of the leaves.

        Has two methods:
            method="average" colors based on the average color of all leaves contained within the tree
            method="complete" colors only if every leaf is of that color, otherwise colors the default_color

        default_color is set to black unless you change it, pass a string or an RGBA tuple
        '''

        for nonterm in self.branches:

            x = []
            for term in nonterm.get_terminals():

                try:
                    x.append(from_branch_color(term.color))
                except:
                    None

            if len(x) > 0:

                if method=='average':
                    color = np.nanmean(x, axis=0)

                elif method=='complete':
                    if np.all(x)==x[0]:
                        color = x[0]
                    else:
                        color = to_rgb(default_color)
            else:
                color = to_rgb(default_color)

            term.color = to_branch_color(color)

    def copy(self):

        '''Return an identical copy of the tree.'''

        return Tree(deepcopy(self._biopython))


    def set_ids(self, new_ids, in_place=True):

        '''Change the ids of the tree, modifying both the leaves' .name attribute and the
            .ids and .leaf_names attributes of the tree.

            One optional modifier: in_place decides if it modifies the tree in place or returns a copy.'''

        if in_place:

            if isinstance(new_ids, (list, np.ndarray)):
                if len(new_ids) == self.N:

                    for leaf,name in zip(self.leaves, new_ids):
                        leaf.name = name

                    self.ids = new_ids
                    self.leaf_names = new_ids

            elif isinstance(new_ids, dict):

                for leaf_id, new_name in new_ids.items():
                    leaf_ind = self.get_leaf_index(leaf_id)
                    self.leaves[leaf_ind].name = new_name
                    self.ids[leaf_ind] = new_name

                self.leaf_names = self.ids

            self._update_ete3_tree()

        else:
            newtree = set_ids(self.copy())

    def draw_ascii(self):

        print(self.ete3.get_ascii())

    def get_ascii(self):

        return self.ete3.get_ascii()

    def save(self, filename, fmt='default'):

        if fmt=='default':

            if self.is_colored:
                fmt="phyloXML"

            else:
                fmt="newick"

        if self.is_colored and fmt=="newick":
            print("Warning: newick format does not store color information, so colors will be lost!")

        Phylo.write(self._biopython, filename, fmt)


def ColorLeaf(leaf, color):

    '''Color a leaf of a tree. Automatically handles a variety of formats:
        - if the input is already a BranchColor, just assign it
        - if it's a string, convert to RGB and then to a BranchColor
        - if it's an RGB or RGBA tuple, convert to BranchColor'''

    if isinstance(color, Phylo.BaseTree.BranchColor):
        leaf.color = color

    elif isinstance(color, (str, np.str_)):
        leaf.color = to_branch_color(to_rgba(color))

    elif len(color)==4 or len(color)==3:
        leaf.color = to_branch_color(color)

    else:
        raise InputError("Must pass either a named color or RGBA values as a list")


def RFdistance(Tree1, Tree2, normalized=False):

    '''Calculate the Robinson-Foulds distance between two trees'''

    rf = Tree1.ete3.robinson_foulds(Tree2.ete3)

    if normalized:
        return rf[0] / rf[1]

    else:
        return rf[0]


def to_branch_color(rgb_color):

    '''Converts an RGB color to a Bio.Phylo BranchColor object'''

    return [int(255*i) for i in rgb_color[:3]]

def from_branch_color(branch_color):

    '''Converts a Bio.Phylo BranchColor object back to an RGB value'''

    return np.array([branch_color.red/255, branch_color.green/255, branch_color.blue/255])

def readTree(filename, fmt="detect", **kwargs):

    '''
    Read a Tree object in from a file.
    '''

    # Autodetect format from filename, or else try newick and phyloXML
    if fmt=="detect":
        try:
            suff = filename.split('.')[-1]

            if suff in ("nwk", "newick"):
                fmt = "newick"

            elif suff in ("xml", "phyloxml"):
                fmt = "phyloxml"

        except:
            print("Unable to autodetect format, guessing newick...")

            try:
                return readTree(filename, fmt="newick")
            except:
                print("Trying phyloXML...")
                return readTree(filename, fmt="phyloXML")

    return Tree(Phylo.read(filename, fmt), **kwargs)

def readTrees(filename, fmt="detect", **kwargs):

    '''
    Read a Tree object in from a file.
    '''

    # Autodetect format from filename, or else try newick and phyloXML
    if fmt=="detect":
        try:
            suff = filename.split('.')[-1]

            if suff in ("nwk", "newick"):
                fmt = "newick"

            elif suff in ("xml", "phyloxml"):
                fmt = "phyloxml"

        except:
            print("Unable to autodetect format, guessing newick...")

            try:
                return readTrees(filename, fmt="newick")
            except:
                print("Trying phyloXML...")
                return readTrees(filename, fmt="phyloXML")

    return Forest([Tree(tree) for tree in Phylo.parse(filename, fmt, **kwargs)])

def read_mrbayes_trprobs(filename):

    '''Read in an ensemble of trees (a "forest") from a MrBayes .trprobs output file.'''

    file = ''

    with open(filename) as f:
        for line in f:
            file = file+line

    ids = {}
    for i in file.split(';')[1].split('\n'):
        l = i.split(' ')
        if len(l) > 4:
            l = [i for i in l if len(i) > 0]
            ids[l[0]] = l[1].replace(',','')

    trees = []; ps = []; Ps = []

    for line in tqdm(file.split(';')[2:]):
        x = line[line.find('['):]
        if 'p = ' in x and ',' in x:
            p = float(x.split("&W ")[1].split(']')[0])
            #P = float(x.split("P = ")[1].split('] = [')[0])
            tree = x.split(' ')[-1]+';'
            tree = Tree(tree)
            tree.set_ids(ids)
            trees.append(tree)
            ps.append(p)
            #Ps.append(P)

    return Forest(trees, np.array(ps), names=list(ids.values()))


class Forest:

    def __init__(self, trees, probs=None, names=None, renormalize=True):

        '''
        Create a Forest object. Takes:

            trees: the input trees, as bioviper.phylo.Tree objects
            probs (optional): probability values associated with each tree, as from a Bayesian tree builder's posterior distribution
            names: a set of names for the nodes; not sure what this is doing at the moment, I might get rid of it but want to make sure it doesn't break stuff
            renormalize: whether to renormalize probabilities to make sure that they sum to one.
                        (this is often a problem when the software outputs them with insufficient significant figures
                         or if you simply pass unnormalized probabilites)
        '''

        self.trees = trees

        if type(probs)==type(None):
            self.p = np.ones(len(self.trees))
        else:
            self.p = probs

        if renormalize and np.sum(self.p) != 1:
            self.p = self.p / np.sum(self.p)

        #self.P = Probs
        self.names = self.trees[0].ids
        self.ids = self.names; self.term_names = self.names
        self.N = len(self.trees)
        #self.consensus_trees = None

    def __getitem__(self, index):
        return self.trees[index]

    def __iter__(self):
        return iter(self.trees)

    def root_all(self, outgroup):

        for tr in self.trees:
            if 'mid' in outgroup:
                root_point = tr.get_midpoint_outgroup()
                tr.set_outgroup(root_point)
            else:
                tr.set_outgroup(outgroup)

    # def find_consensus(self, method="majority"):

    #     if method=="majority":
    #         return Consensus.strict_consensus(self.trees)



    def prune(self, leaves, outgroup=False, rebin=True):


        """Prune the trees to contain only a subset of leaves."""

        rooted=False
        # Prune trees to include only leaves of interest
        pruned_trees = []

        for tree in tqdm(self.trees):
            tr = tree.prune(leaves)

            # if you gave an outgroup
            if type(outgroup) != type(None):
                rooted=True
                if outgroup=='midpoint':
                    root_point = tr.get_midpoint_outgroup()
                    tr.set_outgroup(root_point)
                else:
                    tr.set_outgroup(outgroup)
            pruned_trees.append(tr)

        if rebin:

            unique_trees, counts, x = bin_trees(pruned_trees, rooted=rooted)

            P = []

            for i in x:
                P.append(np.sum(self.p[np.isin(np.arange(self.N), i)]))

            f = Forest(unique_trees, P, np.arange(len(P)))
            f.counts = counts

            return f

        else:
            return Forest(pruned_trees, p=self.p)


    def calc_marginals(self, leaves, outgroup=None):

        """Calculate the marginal probabilities of tree topologies considering only a subset of leaves.
           Takes ETE3 trees collected into a Forest object, presumably loaded from a MrBayes .trprobs file.
           Relies on trees representing a converged MCMC sample with burnout phase removed such that values
           represent true posterior marginal probabilities.

           Parameters:
            --leaves: the names of the leaves of interest, as a list. should be >3 (>2 if rooted) or else
                       there is only one meaningful topology. all leaves must be exactly as in the tree.
            optional:
                --outgroup: identify one of your leaves as an outgroup to do rooted RF distances. can also
                            pass "midpoint", in which case each tree will be midpoint rooted.
                --v: verbose, true displays results in ascii, false only returns probs and topologies"""

            # will change later if you passed an outgroup

        pruned_forest = self.prune(leaves, outgroup=outgroup, rebin=True)

        for k,tree in enumerate(pruned_forest):
            print("Topology",k+1,": p = "+str(pruned_forest.p[k])[:8], "with ", pruned_forest.counts[k], "architectures")
            print(tree.get_ascii())
            print('')
            print('')

        return pruned_forest

def rf_distances(tree_1, tree_list):
    return np.array([RFdistance(tree_1, tree_2) for tree_2 in tree_list])

def bin_trees(forest, rooted=False):

    """Given a set of ETE3 trees, some of which are identical, returns the:
        -- unique trees from the set
        -- counts of each unique tree
        -- indices in the original list corresponding to each unique trees.

    It does so while avoiding calculating the full matrix of robinson-fould distances
    between every pair for efficiency, under the assumption that this matrix will be full of zeros."""

    # calculate the robinson-fould distances of each tree to the first one
    rfdists = rf_distances(forest[0], forest)

    # many of those may be the same; initialize their indices and add them to a growing list
    tree_bins = [np.where(rfdists==0)[0]]

    # these are the indices of trees that are NOT identical, that we now want to consider.
    inds = np.where(rfdists > 0)[0]

    # we will not pay any more attention to the trees matching architecture #1, and only focus on the others now
    new_forest = [forest[i] for i in inds]

    # iterate until we've identified all of the unique architectures
    while len(new_forest) > 0:

        # do basically the same thing
        rfdists = rf_distances(new_forest[0], new_forest)
        tree_bins.append(inds[np.where(rfdists==0)[0]])
        inds = inds[np.where(rfdists>0)[0]]
        new_forest = [forest[i] for i in inds]

    # pick th first tree of each set as its representative - they'll all be identical
    representative_trees = [forest[tree_bins[k][0]] for k in range(len(tree_bins))]
    counts = np.array([len(tree_bins[k]) for k in range(len(tree_bins))])

    return representative_trees, counts, tree_bins



import numpy as np
import pandas as pd
from Bio import AlignIO, Phylo, SeqIO, PDB, pairwise2

from Bio.Align import _aligners
from Bio.Align import substitution_matrices
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord,_RestrictedDict
from Bio.Align import MultipleSeqAlignment

from .pdb import readPDB
from .phylo import readTree, Tree

# Otherwise we get some annoying pandas warnings
import warnings
warnings.simplefilter("ignore", UserWarning)

from copy import copy,deepcopy

alphabet = '-ACDEFGHIKLMNPQRSTVWY'

def readAlignment(alignment_file, format="fasta", calc_frequencies=False):

    '''
    Read in an alignment in a format accepted by biopython. Arguments:
        -alignment_file: the path to the alignment you want to load
        -format: 
    '''

    msa = MultipleSequenceAlignment(AlignIO.read(alignment_file, format=format))

    if calc_frequencies==True:
        msa.calc_frequencies()
        msa.calc_coverage()

    return msa

class MultipleSequenceAlignment(MultipleSeqAlignment):

    """Represents a classical multiple sequence alignment (MSA).

    By this we mean a collection of sequences (usually shown as rows) which
    are all the same length (usually with gap characters for insertions or
    padding). The data can then be regarded as a matrix of letters, with well
    defined columns."""

    def __init__(
        self, records, ids=None, names=None, descriptions=None,
        alphabet=None, annotations=None, column_annotations=None,
        tree=None, _tree_branches=None
    ):
        """Initialize a new MultipleSeqAlignment object.

        Arguments:
         - records - A list (or iterator) of SeqRecord objects, whose
                     sequences are all the same length.  This may be an be an
                     empty list.
         - alphabet - For backward compatibility only; its value should always
                      be None.
         - annotations - Information about the whole alignment (dictionary).
         - column_annotations - Per column annotation (restricted dictionary).
                      This holds Python sequences (lists, strings, tuples)
                      whose length matches the number of columns. A typical
                      use would be a secondary structure consensus string."""

        if alphabet is not None:
            raise ValueError("The alphabet argument is no longer supported")

        # IF you generate from a matrix rather than a list of SeqRecords
        if isinstance(records, np.ndarray):
            self._records = []
            self.matrix = records

            record_list = []

            for k,rec in enumerate(records):

                record_info = []

                seq = SeqRecord(Seq(''.join(rec)))

                if type(ids) != type(None):
                    seq.id = ids[k]
                else:
                    seq.id = None

                if type(names) != type(None):
                    seq.name = str(names[k])

                if type(descriptions) != type(None):
                    seq.description = str(descriptions[k])

                elif type(_tree_branches) != type(None):
                    seq.branch = _tree_branches[k]

                record_list.append(seq)

            super().__init__(record_list)

        else:
            '''If you pass it a list of record objects, it will initialize it as a
            normal MultipleSeqAlignment object.'''

            super().__init__(records)

        # Annotations about the whole alignment
        if annotations is None:
            annotations = {}
        elif not isinstance(annotations, dict):
            raise TypeError("annotations argument should be a dict")
        self.annotations = annotations

        # Annotations about each column of the alignment
        if column_annotations is None:
            column_annotations = {}
        # Handle this via the property set function which will validate it
        self.column_annotations = column_annotations

        #Useful for indexing
        self.N, self.L = self.matrix.shape
        id_to_index = pd.DataFrame(zip(self.ids, np.arange(self.N)), columns=("id", "index")).set_index("id")
        self.id_to_index = id_to_index

        self.coverage = None
        self.frequencies = None
        self.tree = None
        self.numeric_matrix = None
        self.fij = None
        self._pseudocount = 0
        self.sequence_weights = np.ones(self.N)

    def save(self, filename, format='Auto'):

        if format=='Auto':
            try:
                ending = filename.split('.')[-1]
                if "fa" in ending:
                    format = "fasta"
                elif "sto" in ending:
                    format = "stockholm"
                elif "clust" in ending:
                    format = "clustal"
            except:
                return TypeError("Please pass a format or a filename with an ending containing fa, sto or clust")

        AlignIO.write(MultipleSeqAlignment(self._records), filename, format)

    def update(self):
        if self.matrix.shape[0] != len(self._records):
            self.matrix = np.array(self.records)
            self.N, self.L = self.matrix.shape
            self.ids = np.array([rec.id for rec in self._records])
            self.names = np.array([rec.name for rec in self._records])
            self.descriptions = [rec.description for rec in self._records]

    def dealign(self):

        """Return raw sequences without any gap characters (or lowercase)"""

        ds = []

        for rec in self._records:
            s = Seq(str(rec.seq).replace('-','').replace('.','').upper())
            ds.append(SeqRecord(s, id = rec.id, name=rec.name, description=rec.description))

        return SequenceArray(ds)


    def sequence_lengths(self):

        W = (self.matrix!='-')&(self.matrix!='.')
        return np.sum(W,axis=1)

    def calc_coverage(self):
        if type(self.coverage)==type(None):
            self.coverage = np.mean((self.matrix!='.')&(self.matrix!='-'),axis=0)
            return self.coverage
        else:
            return self.coverage

    def calc_frequencies(self, pseudocount=0, first_pos=0, pos_indices = None, store=True):

        self._pseudocount = pseudocount

        M_upper = self.upper().replace('.','-')
        if pseudocount==0:
            frequencies = np.array([ \
                            [np.average(M_upper.matrix[:,k]==i, weights=self.sequence_weights) for i in alphabet] \
                          for k in range(M_upper.matrix.shape[1])])

        else:
            frequencies = np.array([ \
                             [np.average(M_upper.matrix[:,k]==i, weights=self.sequence_weights) + \
                                pseudocount/np.sum(self.sequence_weights) for i in alphabet] \
                            for k in range(M_upper.matrix.shape[1])])
            frequencies = frequencies / np.sum(frequencies, axis=1)[:,None]

        frequencies = pd.DataFrame(frequencies, columns=list(alphabet))

        frequencies.sequence_weights = self.sequence_weights

        if type(pos_indices) != type(None):
            frequencies.index = pos_indices
            
        elif first_pos != 0:
            frequencies.index = np.arange(first_index, self.L + first_index)

        if store:
            self.frequencies = frequencies

        return frequencies

    def as_numeric(self, recalculate=False):

        if type(self.numeric_matrix) == type(None) or recalculate:
            m = copy(self.upper().replace('.','-').matrix)
            mnum = np.zeros(m.shape)
            for n,letter in enumerate(alphabet):
                mnum[m==letter]=n
            self.numeric_matrix = mnum.astype('int')

        return self.numeric_matrix

    def as_df(self, numeric=False):

        '''Return alignment as a pandas dataframe.'''

        if numeric:
            df = pd.DataFrame.from_records(self.as_numeric(), index=self.ids)
        else:
            df = pd.DataFrame.from_records(self.matrix, index=self.ids)

        return df

    def one_hot_encode(self, dim=1):

        '''Generate a one-hot encoded pandas dataframe, e.g. for dimensionality reduction. If dim=1, will be N x L*20, if dim=2, will be N x L x 20'''

        return np.array([one_hot_encode(seq, dim=dim) for seq in self.as_numeric()])


    def calc_pair_frequencies(self, pseudocount=0):

        if type(self.numeric_matrix) == type(None):
            self.as_numeric()

        X = [[[] for i in range(self.L)] for i in range(self.L)]

        for i in range(self.L):
            for j in range(i, self.L):
                A = np.histogram2d(self.numeric_matrix[:,i], self.numeric_matrix[:,j], bins=[np.arange(22), np.arange(22)])[0] + pseudocount
                A = A/np.sum(A)
                X[i][j] = A
                X[j][i] = A

        self.fij = np.array(X)

        return self.fij

    def _set_per_column_annotations(self, value):
        if not isinstance(value, dict):
            raise TypeError(
                "The per-column-annotations should be a (restricted) dictionary."
            )
        # Turn this into a restricted-dictionary (and check the entries)
        if len(self):
            # Use the standard method to get the length
            expected_length = self.get_alignment_length()
            self._per_col_annotations = _RestrictedDict(length=expected_length)
            self._per_col_annotations.update(value)
        else:
            # Bit of a problem case... number of columns is undefined
            self._per_col_annotations = None
            if value:
                raise ValueError(
                    "Can't set per-column-annotations without an alignment"
                )

    def _get_per_column_annotations(self):
        # if self._per_col_annotations is None:
        #     # This happens if empty at initialisation
        #     if len(self):
        #         # Use the standard method to get the length
        #         expected_length = self.get_alignment_length()
        #     else:
        #         # Should this raise an exception? Compare SeqRecord behaviour...
        #         expected_length = 0
        #     self._per_col_annotations = _RestrictedDict(length=expected_length)
        #return self._per_col_annotations

        return None

    column_annotations = property(
        fget=_get_per_column_annotations,
        fset=_set_per_column_annotations,
        doc="""Dictionary of per-letter-annotation for the sequence.""",
    )

    def extend(self, records):
        """Add more SeqRecord objects to the alignment as rows.

        They must all have the same length as the original alignment. For
        example,

        Because the alignment object allows iteration over the rows as
        SeqRecords, you can use the extend method with a second alignment
        (provided its sequences have the same length as the original alignment).
        """
        if len(self):
            # Use the standard method to get the length
            expected_length = self.get_alignment_length()
        else:
            # Take the first record's length
            records = iter(records)  # records arg could be list or iterator
            try:
                rec = next(records)
            except StopIteration:
                # Special case, no records
                return
            expected_length = len(rec)
            self._append(rec, expected_length)
            # Can now setup the per-column-annotations as well, set to None
            # while missing the length:
            self.column_annotations = {}
            # Now continue to the rest of the records as usual

        for rec in records:
            self._append(rec, expected_length)

        try:
            self.matrix = np.concatenate([self.matrix, MultipleSequenceAlignment(records).matrix])
        except:
            self.matrix = np.array(self._records)

        self.ids = np.array([rec.id for rec in self._records])
        self.names = np.array([rec.name for rec in self._records])
        self.descriptions = np.array([rec.description for rec in self._records])

    def upper(self):

        """Make all letters in an alignment uppercase."""

        return MultipleSequenceAlignment(np.char.upper(self.matrix), ids=self.ids, names=self.names, descriptions=self.descriptions)


    def replace(self, a, b):

        """Replace one letter in an alignment with another."""

        return MultipleSequenceAlignment(np.char.replace(self.matrix, a, b), ids=self.ids, names=self.names, descriptions=self.descriptions)

    def standardize(self):

        ''' Make all amino acids upper case and replace all noncanonical characters with gaps ('-') '''

        matrix_upper = np.char.upper(self.matrix)
        noncanonical_characters = [letter for letter in np.unique(matrix_upper) if letter not in alphabet]

        for character in noncanonical_characters:
            matrix_upper = np.char.replace(matrix_upper, character, '-')

        return MultipleSequenceAlignment(matrix_upper, ids=self.ids,
                                names=self.names, descriptions=self.descriptions)

    def set_ids(self, ids):

        """Reset all IDs in place to a new list. Will both change self.ids and the .id attribute
           of each individual sequence record."""

        if len(ids) == self.N:
            self.ids = ids
            for n,seq in enumerate(self._records):
                seq.id = ids[n]
        else:
            raise ValueError('ERROR: cannot use id list of length', len(ids), 'for an alignment with', self.N, 'sequences')


    def modify_ids(melf, func):

        """Modify all IDs in place with a function. Either pre-define the function or use lambda, e.g.
            alignment.modify_ids(lambda x:x.split('|')[2])"""

        ids = []

        for seq in self._records:
            seq.id = func(seq.id)
            ids.append(seq.id)

        self.ids = np.array(ids)

    def pos_covariance(self):

        '''Get the 4D position by position covariance matrix, considering one-hot encoded sequences'''

        return np.cov(self.one_hot_encoding(full=True).T).reshape((sampler.alignment.L, sampler.alignment.L, 21, 21))

    def seq_covariance(self):

        '''Get the covariance matrix across sequences'''

        return np.cov(np.array(self.one_hot_encoding()))

    def __getitem__(self, index):
        """Access part of the alignment. Indexes like a 2D numpy array and returns
        another MultipleSequenceAlignment object unless:
            1) you ask for a single row, when you get a SeqRecord object
            2) you ask for a single column, when you get a numpy array
            3) you ask for a single row AND a single column, when you get a single letter as a string."""

        if isinstance(index, int):
            return self._records[index]

        elif isinstance(index, slice):
            return self.__getitem__(np.arange(self.N)[index])

        # if you pass an array as the index
        elif isinstance(index, np.ndarray):

            # if the first item is an integer, float or boolean
            if isinstance(index[0], (int, np.int64, float, np.float64, bool, np.bool_)):

                ids = self.ids[index]; names = self.names[index]
                descs = list(np.array(self.descriptions)[index])

                # taking the appropriate branches of the tree if you have an attached tree
                if type(self.tree) != type(None):

                    if type(index[0])==np.bool_:
                        branches = [self._records[k] for k in np.where(index)[0]]
                    else:
                        branches = [self._records[k] for k in index]

                    new_msa = MultipleSequenceAlignment(self.matrix[index], ids=ids,
                        names=names, descriptions=descs,_tree_branches=branches)
                    new_msa.tree = self.tree

                    return new_msa

                else:
                    return MultipleSequenceAlignment(self.matrix[index], ids=ids, names=names, descriptions=descs)

            # if they're not floats/integers/boolean 
            else:

                return self.__getitem__(np.array(self.id_to_index.loc[index]["index"]))

        elif isinstance(index, list):
            return self.__getitem__(np.array(id))

        elif isinstance(index, str):
            return self.__getitem__(np.where(self.ids==index)[0])[0]

        elif len(index) != 2:
             raise TypeError("Invalid index type.")

        else:
            # Handle double indexing
            row_index, col_index = index

            # If you ask for a single row...
            if isinstance(row_index, int):

                # and a single, column, it will return a string (same as indexing self.matrix)
                if isinstance(col_index, int):
                    return self.matrix[row_index, col_index]

                # and a slice, return a single record only at that particular set of positions
                elif isinstance(col_index, (slice, np.ndarray)):
                    return self.__getitem__((np.arange(self.N), col_index))._records[row_index]

            # If you pass a single column, it will return an array, not an MSA object with a single column because that's dumb
            #   here is one difference from a MultipleSeqAlignment in Biopython, where a string is returned
            #   it seems to me that an array is a much more logical data structure for this

            elif isinstance(col_index, int):
                return self.matrix[:,col_index]

            # When you pass complicated indexing
            else:

                ids = self.ids[row_index]
                names = self.names[row_index]
                descs = list(np.array(self.descriptions)[row_index])
                #new_msa = MultipleSequenceAlignment(self.matrix[row_index, col_index], ids=ids)

                # if there's an attached tree, copy over the tree without re-attaching
                if type(self.tree) != type(None):
                    new_msa = MultipleSequenceAlignment(self.matrix[row_index, col_index], ids=ids,
                        names=names, descriptions=descs,
                        _tree_branches=[rec.branch for rec in self._records])
                    new_msa.tree = self.tree

                    return new_msa

                else:
                    return MultipleSequenceAlignment(self.matrix[row_index, col_index],
                        ids=ids, names=names, descriptions=descs)


    def search_id(self, id_str):

        """Search for an incomplete ID in the alignment"""

        in_id = np.array([str(id_str) in id_val for id_val in self.ids])
        matched = np.where(in_id)[0].astype('int')

        if len(matched) == 0:
            return None

        elif len(matched) == 1:
            return self.__getitem__(int(matched[0]))

        else:
            return self.__getitem__(matched)

    def search_ids(self, ids):

        seqs = []

        for id_val in ids:
            matched = self.search_id(id_val)

            if not isinstance(matched, type(None)):

                if isinstance(matched, MultipleSequenceAlignment):
                    for match in matched:
                        seqs.append(match)

                elif isinstance(matched, SeqRecord):
                    seqs.append(matched)


        return MultipleSequenceAlignment(seqs)


    def search_sequence(self, sequence, include_gaps=False, case_sensitive=False):


        """Search for a particular piece of sequence in the alignment. Only matches the sequence exactly."""

        x = []
        for nr,record in enumerate(self._records):
            if include_gaps==True:
                if case_sensitive:
                    seq = record.seq
                else:
                    seq = record.seq.upper()
                    sequence=sequence.upper()

                if sequence in seq:
                    x.append(nr, seq.index(sequence))

            else:
                if case_sensitive:
                    seq = record.seq.replace('-','').replace('.').upper()
                else:
                    seq = record.seq.replace('-','').replace('.','').upper()
                    sequence=sequence.upper()

                if sequence in seq:
                    where_ungapped = np.where((self.matrix[nr]!='-')&(self.matrix[nr]!='.'))[0]
                    x.append(nr, where_ungapped[seq.index(sequence)])

        if len(x)==0:
            return None

        if len(x)==1:
            return x[0]

        else:
            return x

    ### TREE FUNCTIONALITY ##

    def attach_tree(self, tree, format='newick', prune_unmatched=False):

        """Attach a precalculated tree to the alignment. See documentation for how usage.
            As of now, IDs in the tree must exactly match IDs in the alignment."""

        if type(tree)==str:
            tree = readTree(tree, format)

        if isinstance(tree, (Phylo.Newick.Tree, Phylo.PhyloXML.Phylogeny)):
            tree = Tree(tree)

        self.tree = deepcopy(tree)

        for leaf in self.tree.leaves:

            try:
                n = list(self.ids).index(leaf.name)

            except:

                ns = [n for n,i in enumerate(self.ids) if leaf.name in i]
                if len(ns) == 0:
                    print("Terminal name", leaf.name, "not in alignment - will not be attached!")
                    n = -1

                elif len(ns) == 1:
                    n = n[0]

                else:
                    print("Ambiguity with terminal time", leaf.name, "in alignment", len(ns), "times!")
                    n = -1

            if n > -1:
                self._records[n].branch = leaf
                leaf.seq = self._records[n].seq
                leaf.nseq = n

            elif prune_unmatched:
                self.tree.prune(leaf)

    def set_sequence_weights(self, weights):

        self.sequence_weights = np.array(weights)
        for n,record in enumerate(self._records):
            record.weight = weights[n]

    def get_index(self, id):

        """Return the index of a particular sequence id. Should work for partial IDs;
            for indices that appear multiple times it will return an array, and for missing
            indices it will return None."""


        try:
            # pandas indexing here is by far the fastest way to do this
            return self.id_to_index.loc[id].iloc[0]

        except:

            a = self.search_id(id)
            if type(a)==type(None):
                return None

            elif type(a) == SeqRecord:
                return self.get_index(a.id)

            else:
                return np.array([self.get_index(seq.id) for seq in a])

    def fix_tree(self):

        """This updates the current attached tree to (1) make sure that the number matching is correct
        and (2) remove any nodes that are not present in the alignment.

        Technically, it makes a copy of the original tree and re-attaches it before modifying.

        You'll need to do this if you subset particular sequences in an alignment with an attached tree
        and now want to use the new tree. I don't do it by default because for very large alignments
        I suspect it may take a little while and maybe you don't care."""

        if type(self.tree) != type(None):
            new_tree = deepcopy(self.tree)
            self.attach_tree(new_tree)
            branches = [record.branch for record in self._records]

            for term in self.tree.get_terminals():
                if tree not in branches:
                    self.tree.prune(branch)


    def subset_by_ids(self, ids, sort=False, match_order=True):

        """
        Subset the MSA using a set of IDs. Used to be implemented in a way that turned out to be much slower, with two
        deprecated arugments (sort & match_order); I've kept them defined by default here just for backwards compatibility
        """

        return self.__getitem__(np.array(self.id_to_index.loc[ids]["index"]))


    def subset_by_clade(self, clade):

        """For a clade in the attached tree, subset the alignment to get only those sequences."""

        try:
            return self.__getitem__(np.array([term.nseq for term in clade.get_terminals()]))

        # This will happen if some of the terminals don't have a .nseq attribute
        except AttributeError:

            # Define the terminals and their names
            terminals = clade.get_terminals()
            terminal_names = [term.name for term in terminals]

            # Find which ones are actually in
            isin = np.isin(terminal_names, self.ids)

            # If any were found, return them
            if np.sum(isin) > 0:
                return self.__getitem__(np.array([terminals[k].nseq for k in np.where(isin)[0]]))

            # Else nothing is returned
            else:
                return None

    def build_tree(self, method='nj', model='autodetect', root=True, ladderize=True, filename="fasttree.fa"):

        "Build a tree and attach it. Calls functions from the TreeBuilders.py script"""

        if method=='nj':
            self.attach_tree(NJTree(MultipleSeqAlignment(self.upper()._records), matrix=model, root=root, ladderize=ladderize))

        elif method=='upgma':
            self.attach_tree(UPGMATree(self, matrix=model, root=root, ladderize=ladderize))

        elif method=='FastTree':
            if model=='autodetect':
                model = "LG"
            format_id = filename.split(".")[-1]
            if "fa" in format_id:
                format = "fasta"
            elif "sto" in format_id:
                format = "stockholm"
            elif "clust" in format_id:
                format = "clustal"
            else:
                print("Cannot interpret file format for", filename, "defaulting to fasta")
                format = "fasta"

            self.attach_tree(FastTree(self, model=model, root=root, ladderize=ladderize, cat=cat,
                                temp_aln=filename))

    def attach_structure(self, structure, nseq=None, name=None, chains='A', model=0):

        if isinstance(structure, str):
            if type(name) == type(None):
                if type(nseq) != type(None):
                    name = self.ids[nseq]
                else:
                    name = ''

            structure = readPDB(structure, name, chains, model)

        if type(nseq) == None:

            x = self.search_sequence(structure.sequence)

            if isinstance(x, NoneType):
                print('Could not find sequence')
                return None

            elif isinstance(x, int):
                nseq = x

            elif isinstance(x, list):
                print(len(x), 'matching sequences found! Appending to first...')
                nseq = x[0]

        elif type(nseq) == str:

            a = np.where(self.ids==nseq)[0]
            if len(a)==0:
                a = self.get_index(nseq)
                if type(a) == type(None):
                    print("string", nseq, "is not located anywhere in alignment IDs!")
            else:
                a = a[0]

            nseq = a

        record = self._records[nseq]

        # Align the structure's sequence to the record's sequence
        #  replace gaps in the record's sequence with "x" to not confuse these
        #  gaps (which we want to remember) with possible new gaps introduced
        #  in the alignment
        A = PairwiseAlign(str(record.seq).replace('-','x'), structure.ordered_sequence,
                            alignment_method='globalms', alignment_params=(2,-1,-0.75,-.1))

        A1 = A[:,np.where(A.matrix[0]!='-')[0]]
        A2 = A[:,np.where(A.matrix[1]!='-')[0]]

        structure._alipos = np.where(A1.matrix[1]!='-')[0]
        structure._pdbpos = np.where(A2.matrix[0]!='-')[0]

        try:
            record.structures
        except:
            record.structures = {}

        if type(name) != None:

            if 0 in record.structures:
                record.structures[np.max([key for key in record.structures.keys() if type(key)==int])+1] = structure
            else:
                record.structures[0] = structure

        else:
            record.structures[name] = structure

        try:
            self.structures.append((nseq, structure))
        except:
            self.structures = [(nseq, structure)]

def readSequences(filename, format="Detect"):

    if format=='Detect':
        ending = filename.split('.')[-1]
        if 'fastq' in ending:
            format="fastq"
        elif 'fa' in ending:
            format="fasta"
        elif 'sto' in ending:
            format = "stockholm"
        elif 'nex' in ending:
            format = "nexus"
        elif "a2m" in ending:
            format = "fasta"

    return SequenceArray(list(SeqIO.parse(filename, format)), fmt=format)

class SequenceArray:

    def __init__(self, records, fmt="fasta"):

        self._records = records
        self.ids = np.array([record.id for record in self._records])
        self.names = np.array([record.id for record in self._records])
        self.descriptions = [record.id for record in self._records]
        self.N = len(self._records)
        self.L = np.array([len(rec.seq) for rec in self._records])
        self.format = fmt

        # For fastq sequencing reads with phred quality scores, save those phred quality scores in a centralized place
        if fmt == "fastq":
            self.phred_quality = [np.array(rec.letter_annotations["phred_quality"]) for rec in self._records]
            self.mean_phred_quality = np.array([np.mean(qual_arr) for qual_arr in self.phred_quality])

    def __len__(self):
        return self.N

    def __getitem__(self, index):

        if isinstance(index, int):
            return self._records[index]

        elif isinstance(index, slice):
            return SequenceArray(self._records[index])

        elif isinstance(index, (list, np.ndarray)):

            if type(index[0])==np.bool_:
                return SequenceArray([self._records[k] for k in np.where(index)[0]])

            else:
                return SequenceArray([self._records[k] for k in index])

        elif isinstance(index, str):

            if index in self.ids:
                return self._records[self.ids.index(index)]
            else:
                raise IndexError("No such ID in SequenceArray. If ID is incomplete, try seqs.search_id(name)")

    def from_file(self, filename, format="Detect"):

        if format=='Detect':
            ending = filename.split('.')[-1]
            if 'fa' in ending:
                format="fasta"
            elif 'sto' in ending:
                format = "stockholm"
            elif 'nex' in ending:
                format = "nexus"
            elif "a2m" in ending:
                format = "fasta"

        self._records = list(SeqIO.parse(filename, format))
        self.ids = np.array([record.id for record in self._records])
        self.names = np.array([record.id for record in self._records])
        self.descriptions = [record.id for record in self._records]

    def search_id(self, id_str):

        """Search for an incomplete ID in the alignment"""

        in_id = np.array([str(id_str) in id_val for id_val in self.ids])

        if np.sum(in_id) > 1:
            return self.__getitem__(np.where(in_id)[0].astype('int'))
        elif np.sum(in_id)==1:
            return self.__getitem__(np.where(in_id)[0].astype('int'))[0]
        else:
            return []

    def __iter__(self):
        return __iter__(self._records)

    def search_ids(self, ids):

        return SequenceArray([self.search_id(id_val) for id_val in ids])

    def search_sequence(self, sequence):

        x = []
        for nr,record in enumerate(self._records):

            if sequence in record.seq:
                x.append(nr, seq.index(sequence))

        if len(x)==0:
            return None

        if len(x)==1:
            return x[0]

        else:
            return x

    def attach_structure(self, structure, nseq=None, name=None, chains='A', model=0):

        if isinstance(structure, str):
            structure = readPDB(filename, name, chains, model)

        if type(nseq) == None:

            x = self.search_sequence(structure.sequence)

            if isinstance(x, NoneType):
                print('Could not find sequence')
                return None

            elif isinstance(x, int):
                nseq = x

            elif isinstance(x, list):
                print(len(x), 'matching sequences found! Appending to first...')
                nseq = x[0]

        elif type(nseq) == str:

            a = np.where(self.ids==nseq)[0]
            if len(a)==0:
                a = self.search_id(nseq)[0]
                if len(a)==0:
                    print("string", nseq, "is not located anywhere in alignment IDs!")
                else:
                    a = a[0]
            else:
                a = a[0]

            nseq = a

        record = self._records[nseq]

        # Align the structure's sequence to the record's sequence
        #  replace gaps in the record's sequence with "x" to not confuse these
        #  gaps (which we want to remember) with possible new gaps introduced
        #  in the alignment
        A = PairwiseAlign(str(record.seq).replace('-','x'), structure.ordered_sequence,
                            alignment_method='globalms', alignment_params=(2,-1,-0.75,-.1))

        A1 = A[:,np.where(A.matrix[0]!='-')[0]]
        A2 = A[:,np.where(A.matrix[1]!='-')[0]]

        structure._seqpos = np.where(A1.matrix[1]!='-')[0]
        structure._pdbpos = np.where(A2.matrix[1]!='-')[0]

        try:
            record.structures
        except:
            record.structures = {}

        if type(name) != None:

            if 0 in record.structures:
                record.structures[np.max([key for key in record.structures.keys() if type(key)==int])+1] = structure
            else:
                record.structures[0] = structure

        else:
            record.structures[name] = structure

        try:
            self.structures.append((nseq, structure))
        except:
            self.structures = [(nseq, structure)]

    def save(self, filename, format='Auto'):

        if format=='Auto':
            try:
                ending = filename.split('.')[-1]
                if "fa" in ending:
                    format = "fasta"
                elif "sto" in ending:
                    format = "stockholm"
                elif "clust" in ending:
                    format = "clustal"
            except:
                return TypeError("Please pass a format or a filename with an ending containing fa, sto or clust")

        SeqIO.write(self._records, filename, format)


def PairwiseAlign(sequence1, sequence2, alignment_method, alignment_params):

    '''
    Pairwise alignment of two sequences using simple alignment tools from pairwise2.align. A standalone function
    to make it easier to incorporate choices of different methods or parameters into a more complex workflow
    '''

    if alignment_method == 'globalms':
        alignment = pairwise2.align.globalms(sequence1, sequence2, *alignment_params)[0]
    elif alignment_method == 'globalxx':
        alignment = pairwise2.align.globalxx(sequence1, sequence2, *alignment_params)[0]
    elif alignment_method == 'globalmx':
        alignment = pairwise2.align.globalmx(sequence1, sequence2, *alignment_params)[0]
    elif alignment_method == 'localxx':
        alignment = pairwise2.align.localxx(sequence1, sequence2, *alignment_params)[0]
    elif alignment_method == 'muscle':
        alignment = MuscleAlign(sequence1, sequence2, muscle_path)
    else:
        print('Not a valid option for alignment! Currently supported options are globalxx, globalmx, globalms, and localxx')

    ali = np.array([np.array(list(alignment[0])), np.array(list(alignment[1]))])
    return MultipleSequenceAlignment(ali)


def cluster_sizes(ali, thresh):

    try:
        imat = ali.identity_matrix
    except:
        imat = CalcIdentityMatrix(ali)

    return np.array([np.sum(imat[i] > thresh) for i in range(ali.N)])

def CalcIdentityMatrix(msa, gap_handling=1):

    '''Code to calculate a percent identity matrix between every pair of sequences
    in a multiple sequence alignment. Uses numpy array broadcasting and so is much faster
    than Biopython's built-in method.

    Comes with three options for handling gap characters when calculating identity:
        gap_handling = 0 ("keep_all") :
                            treats gaps as a character, so two gaps are "equal"

        gap_handling = 1 ("ignore_doubles") :
                            doesn't include positions where both sequences are gapped but
                            if only one is gapped this counts as a difference
                            (DEFAULT, this one makes the most sense to me)

        gap_handling = 2 ("ignore_all"):
                            ignore positions that are gapped in either sequence
    '''

    # Initialize the identity matrix
    IM = []
    for nseq, seq in enumerate(msa.matrix):

        # If we're treating gaps as a character, this is easy
        #  loop through sequences first rather than doing it all at once with numpy to lower memory usage
        if str(gap_handling) in ('0', 'keep_all', 'False'):

            # Take the mean identity of the sequence with every other sequence
            IM.append(np.mean(seq==msa.matrix, axis=1))

        # We need to do a little bit more to ignore gaps
        if str(gap_handling) in ('1', 'ignore_doubles'):

            # We're going to loop through sequences again
            ids = []

            # Don't need to precalculate the ._coverage_mask if it's already defined for that MSA
            try:
                msa._coverage_mask[nseq]
            except:
                msa._coverage_mask = np.array([msa.matrix[n]!='-' for n in range(msa.N)])

            for n in range(msa.N):

                # If either is TRUE, return TRUE
                use_pos = np.add(msa._coverage_mask[n], msa._coverage_mask[nseq])

                #
                ids.append(np.mean(seq[use_pos]==msa.matrix[n][use_pos]))

            IM.append(ids)

        if str(gap_handling) in ('2', 'ignore_all'):

            ids = []

            try:
                msa._coverage_mask[nseq]
            except:
                msa._coverage_mask = np.array([msa.matrix[n]!='-' for n in range(msa.N)])

            for n in range(msa.N):

                # If either is FALSE, return FALSE
                use_pos = msa._coverage_mask[n] & msa._coverage_mask[nseq]

                ids.append(np.mean(seq[use_pos]==msa.matrix[n][use_pos]))

            IM.append(ids)

    return np.array(IM)


def one_hot_encode(seq_num, dim=1):

    '''
    One-hot encode a numerically-encoded amino acid sequence. Faster than training a one-hot encoder
    and using .transform()
    '''

    L = len(seq_num)

    # Initialize an array of all zeros of the correct shape
    x = np.zeros((L,21))

    # Assign the appropriate positions to be ones
    x[np.arange(L), seq_num] = 1

    # Make one-dimensional if dim=1
    if dim==1:
        return x.reshape((L*21))
    elif dim==2:
        return x



from .msa import MultipleSequenceAlignment, SequenceArray
from .msa import readAlignment, readSequences, PairwiseAlign, CalcIdentityMatrix, cluster_sizes

from .pdb import ProteinStructure
from .pdb import readPDB, readAlphaFold, CalcDistanceMatrix

#from .TreeBuilders import NJTree, UPGMATree, MuscleAlign, FastTree



from __init__ import *
import subprocess
import Bio

def hmmalign(sequences, hmm, outfilename='tmp.sto', reload=True):

    if type(sequences)==str:
        seqfilename = sequences
    else:
        Bio.SeqIO.write(sequences, "tmpseqs.fasta", "fasta")
        seqfilename = "tmpseqs.fasta"

    if type(hmm)==str:
        hmmfilename = hmm
    elif isinstance(hmm, MultipleSequenceAlignment):
        msa = hmm
        hmmbuild(msa, "tmp.hmm")
        hmmfilename = "tmp.hmm"

    cmd = "hmmalign -o "+outfilename+" " + hmmfilename + " " + seqfilename
    subprocess.call("module load hmmer;"+cmd, shell=True)

    if reload==True:
        ali = readAlignment(outfilename, "stockholm")
        return ali

def hmmbuild(msa, hmm_filename, aliname="tmp.fa"):

    msa.save(aliname)
    subprocess.call("module load hmmer; hmmbuild " + hmm_filename + " "  + aliname, shell=True)

#def reformat(input_filename,

    # Reformat an alignment using esl-reformat without explicitly loading it into memory -
        # super fast

default_hmmer_bin = "/n/gaudet_lab/programs/hmmer-bin/"

def esl_alimask(alignment, save_as='tmp.sto', output = "Default", hmmer_dir=default_hmmer_bin,
               p = True, ppcons = True, ppcons_thresh = 0.5, g = False, gapthresh = 0.5,
               final_format = "stockholm", load = True):

    """ This serves as a wrapper function to call the esl tool alimask, which is great at very quickly subsetting columns
        of a huge alignment based on gap frequencies or posterior probability values. While these are all possible with the
        MultipleSequenceAlignment object, doing so for alignments with many columns is slow and requires a huge amount of
        memory, while this can do so in seconds.

        By default, this function masks based on a 50% consensus probability threshold, but can support a variety of other
        functionalities:
            p = True/False (Default: TRUE) controls whether a probability threshold is used
            ppcons = True/Fase (Default: TRUE) controls whether the probability threshold is based on the consensus probability
                annotation for each column.
            ppcons_thresh = <float> (0-1) (Default: 0.5) sets the consensus probability threshold.

            g = True/False (Default: FALSE): masks by fraction of gaps, either in addition to or instead of the probability threshold.
            gapthresh = <float> (Default: 0.5) sets the gap fraction, between 0 and 1

        """

    if type(alignment)==str:
        alifilename = sequences

    else:
        alignment.save(save_as, "stockholm")
        alifilename = save_as

    if output == "Default":
        output = ''.join(alifilename.split('.')[:-1]) + '_masked.sto'

    cmd = hmmer_dir + os.sep + "esl-alimask -o " + output

    if p:
        cmd = cmd + ' -p'

        if ppcons:
            cmd = cmd + ' --ppcons ' + str(ppcons_thresh)

    if g:
        cmd = cmd + ' -g'

        if gap_thresh != 0.5:
            cmd = cmd + ' --gapthresh ' + str(gapthresh)

    subprocess.call(cmd, shell=True)

    if final_format != 'stockholm':

        ali = reformat(output, "afa")

        if load:
            return ali

    elif load:
        return readAlignment(output, "stockholm")



from distutils.core import setup
setup(name='bioviper',
      version='0.2.0',
      author='Sam Berry',
      author_email="sberry@g.harvard.edu",
      description = "Enhancements to Biopython for working with biological data",
      py_modules=["bioviper", "bioviper.msa", "bioviper.pdb", "bioviper.hmmer_tools",
                  "bioviper.phylo"]
      )
